
/* 
Include Files
*/
#include <Blackfin_definitions.h>
#include <functions.h>

/*
Global Variables
*/

char CommandBuffer[6][8];
char ResponseBuffer[17][8];
char FinalResponseBuffer[17];
char TransmitData[516][8];
char ReceiveData[517][8];
char FinalReceiveData[514];
int InterruptCause;
int TimerInterruptCause;
int ClockDone;
int PPIDone;
int ReadyForInput;
char MaxSpeed;
char MaxReadBlkLen;
int CardSize;
char Busy;
char SectorSize;
char MaxWriteBlkLen;
int NewRCA;


/*
Function Definitions
*/

void PLLSetupBF()
{
    *pSIC_IWR = 0x1;
	*pPLL_CTL = SET_MSEL(VCO / CLKin);
	ssync();
	idle();
	*pPLL_DIV = SET_SSEL(VCO / SystemCLK) | ((VCO_CoreCLK)&0x30);
	asm("ssync;");
}

void PortMuxingBF()
{
	*pPORTF_FER = PF0 | PF1 | PF2 | PF3 | PF4;
	*pPORTG_FER = PG10 | PG11;
	*pPORTG_MUX |= 0x1000;
}

void SetUpInterruptsBF()
{
    register_handler(ik_ivg8,MMCIsr);
    *pILAT = 0;
    *pSIC_IMASK0 |= IRQ_DMA0;
    register_handler(ik_timer,TimerISR);
    register_handler(ik_ivg12,PFIsr);
    *pSIC_IAR5 &= 0xFF5FFFFF;
    *pSIC_IAR5 |= 0x500000;
    register_handler(ik_ivg13,PFDataIsr);
}

void MMCInit()
{
	char tempvar;
    int i = 0;
    SetupPFpin();
    
    ReadyForInput = 0;
    SetUpClockSD(8);    //8 cycles needed to initialize the SD Card
    while(ClockDone);
    
    SetUpClockSDContinuous();
    SendCommand(0,0);
    
    SetUpClockSD(8);    //8 cycles to meet the SD card Spec
    while(ClockDone);
    
    SetUpClockSDContinuous();
    do
    {
    	SendCommand(55,0);  	/*Tx - Send CMD 55 - Expect R1 response */
        ReceiveResponse(48);  	/*Rx - Receive R1 response */
                        
        SetUpClockSD(8);    	//8 cycles to meet the SD card Spec
		while(ClockDone);
    
		SetUpClockSDContinuous();
        
	    SendCommand(41, 0x00FF8000);
	    ReceiveResponse(48);  	/*Rx - Receive R3 response */
	    ExtractResponse();
	    
	    SetUpClockSD(8);    	//8 cycles to meet the SD card Spec
		while(ClockDone);
    
		SetUpClockSDContinuous();
    	
		//This delay is given to meet the SD Card Spec
		SetUpClockSD(75000);   
	    while(ClockDone);	   
    
	    SetUpClockSD(75000);   
	    while(ClockDone);	   
    
	    tempvar = FinalResponseBuffer[1] & 0x80;
    } while ((tempvar & 0x80) != 0x80);
    
    
    
    SendCommand(2,0);  /*Tx - Send CMD 2 - Expect R2 response */
    ReceiveResponse(136);  /*Rx - Receive R2 response */
    
    
    SetUpClockSD(8);    //8 cycles to meet the SD card Spec
	while(ClockDone);
    
	SetUpClockSDContinuous();
    
	SendCommand(3,0);  /*Tx - Send CMD 3 - Expect R6 response */
    ReceiveResponse(48);  /*Rx - Receive R6 response */
    SetUpClockSDContinuous();
    ExtractResponse();
    
    SetUpClockSD(8);    //8 cycles to meet the SD card Spec
	while(ClockDone);
    
	SetUpClockSDContinuous();
    
    NewRCA = (((FinalResponseBuffer[1]<<8) & 0xFF00) | ((FinalResponseBuffer[2]) & 0xFF));
    
	SendCommand(9,((NewRCA<<16) & 0xFFFF0000));  /*Tx - Send CMD 9 - Expect R2 response */
    ReceiveResponse(136);  /*Rx - Receive R2 response */
    SetUpClockSDContinuous();
    ExtractResponse();
    
    SetUpClockSD(8);    //8 cycles to meet the SD card Spec
	while(ClockDone);
    
	SetUpClockSDContinuous();
    
    MaxSpeed = FinalResponseBuffer[4];
    MaxReadBlkLen = (FinalResponseBuffer[6] & 0x0F);
    CardSize = (((FinalResponseBuffer[7] & 0x3)<<10) | (FinalResponseBuffer[8] << 2) | (FinalResponseBuffer[9] & 0xC0));
    SectorSize = (((FinalResponseBuffer[11] & 0x3F)<<1) | ((FinalResponseBuffer[12] & 0x80)>>7));
    MaxWriteBlkLen = (((FinalResponseBuffer[13] & 0x3)<<2) | ((FinalResponseBuffer[14] & 0xC0)>>6));  
    
    SendCommand(7,((NewRCA<<16) & 0xFFFF0000));  /*Tx - Send CMD 7 - Expect R1b response */
    ReceiveResponse(48);  /*Rx - Receive R1b response */
    
    SetUpClockSD(8);    //8 cycles to meet the SD card Spec
	while(ClockDone);
    
	SetUpClockSDContinuous();
    
}

void SingleBlockWrite()
{
    int i;
    SetUpTransmitData();
    
    SendCommand(16,0);  //Tx - Send CMD 16 - Expect R1 response 
    ReceiveResponse(48);  //Rx - Receive R1 response 
    SetUpClockSDContinuous();
       
    SetUpClockSD(8);    //8 cycles to meet the SD card Spec
	while(ClockDone);
    
	SetUpClockSDContinuous();
    
    SendCommand(24,0);  //Tx - Send CMD 24 - Expect R1 response
    ReceiveResponse(48);  //Rx - Receive R1 response 
    SetUpClockSDContinuous();
   
    SetUpDMADataTransmit(4121);
    SetUpAndEnablePPITransmit(4120);
        
    SetUpClockBF(4); //Few clock cycles given only to BF to account for the delay
    while(ClockDone);
    
    SetUpClock(4121);
    while(ClockDone);
    while(PPIDone);
    
    InterruptCause = 2; //Dummy DMA transaction
    SetUpDMAReceive(8);
    SetUpAndEnablePPIReceive();
    SetupPFInterruptData();  //Routine to sense the first zero sent by the card
    
    SetUpClockSDContinuous();
	while(!ReadyForInput);
    ReadyForInput = 0;
    
    SetUpClockBFContinuous();
    while(PPIDone);
    
    InterruptCause = 3; //Actual DMA to receive the response
    SetUpDMADataReceive(8);
    SetUpClockBFReceive();
    while(ClockDone);
    SetUpClock(8);
    while(PPIDone);
    while(ClockDone);
    
    SetUpClockSDContinuous();
    ExtractData();
    if(FinalReceiveData[0] != 0x28)
    {
    	while(1);
    	//Error in write - CRC check failed
    }
    
    Busy = 0;
    
    *pPORTFIO_INEN = PF6;
    Busy = ((*pPORTFIO & PF6)>>6);
    while(!Busy)
    {
    	Busy = ((*pPORTFIO & PF6)>>6);
    }
    Busy = 0;
       
}

void SingleBlockRead()
{    
    SendCommand(16,512);  //Tx - Send CMD 16 - Expect R1 response 
    ReceiveResponse(48);  //Rx - Receive R1 response 
    
    SetUpClockSDContinuous();
       
    SetUpClockSD(8);    //8 cycles to meet the SD card Spec
	while(ClockDone);
    
	SetUpClockSDContinuous();
    
    SendCommand(17,0);  /*Tx - Send CMD 17 - Expect R1 response */
    
    SetupPFInterrupt();  //Routine to sense the first zero sent by the card
    SetUpClockSDContinuous();
    while(!ReadyForInput);
    ReadyForInput = 0;  
    
    InterruptCause = 2; //Dummy DMA transaction
    SetUpDMAReceive(8);
    SetUpAndEnablePPIReceive();
    SetUpClockBFContinuous();
    while(PPIDone);
    
    InterruptCause = 3; //Actual DMA to receive the response
    SetUpDMADataReceive(4160);
    SetUpClockBFReceive();
    while(ClockDone);
    SetUpClock(4160);
    while(PPIDone);
    while(ClockDone);
    
    SetUpClockSDContinuous();
    ExtractRData();
}

void SendCommand(int Command, int value)
{
    SetUpCommandArray(Command, value);
    SetUpDMATransmit(48);
    SetUpAndEnablePPITransmit(47);
    
    SetUpClockBF(4); //Few clock cycles given only to BF to account for the delay
    while(ClockDone);
    
    SetUpClock(48);
    while(ClockDone);
    while(PPIDone);     
}

void ReceiveResponse(int Type)
{    	
    SetupPFInterrupt();  //Routine to sense the first zero sent by the card
    SetUpClockSDContinuous();
    while(!ReadyForInput);
    ReadyForInput = 0;  
    
    InterruptCause = 2; //Dummy DMA transaction
    SetUpDMAReceive(8);
    SetUpAndEnablePPIReceive();
    SetUpClockBFContinuous();
    while(PPIDone);
    
    InterruptCause = 3; //Actual DMA to receive the response
    SetUpDMAReceive(Type);
    SetUpClockBFReceive();
    while(ClockDone);
    SetUpClock(Type);
    while(PPIDone);
    while(ClockDone);
}


void SetUpClockSD(int NumberOfCycles)
{
    TimerInterruptCause = 1;
    ClockDone = 1;
    
    *pTIMER_DISABLE = TIMDIS7;
	while(*pTIMER_STATUS & TRUN7);
	
    *pTCNTL = TMPWR;
    *pTCOUNT = (NumberOfCycles-1)* ((VCO/VCO_CoreCLK)/TimerSlow);
    *pTSCALE = 0;
       
	*pTIMER7_WIDTH = SystemCLK / TimerSlow / 2;
	*pTIMER7_PERIOD = SystemCLK / TimerSlow;
	*pTIMER7_CONFIG = PERIOD_CNT | PWM_OUT;

	*pTIMER_ENABLE = TIMEN7;
 	*pTCNTL |= TMREN;
}

void SetUpClockSDContinuous()
{
    *pTIMER_DISABLE = TIMDIS7;
	while(*pTIMER_STATUS & TRUN7);
	
    *pTIMER7_WIDTH = SystemCLK / TimerSlow / 2;
	*pTIMER7_PERIOD = SystemCLK / TimerSlow;
	*pTIMER7_CONFIG = PERIOD_CNT | PWM_OUT;

	*pTIMER_ENABLE = TIMEN7;
}

void SetUpClockBFContinuous()
{
    *pTIMER6_WIDTH =  SystemCLK / TimerSlow / 2;
	*pTIMER6_PERIOD = SystemCLK / TimerSlow;
	*pTIMER6_CONFIG = PERIOD_CNT | PWM_OUT;

	*pTIMER_ENABLE = TIMEN6;
}

void SetUpClockBF(int NumberOfCycles)
{
    TimerInterruptCause = 2;
    ClockDone = 1;
	
    *pTCNTL = TMPWR;
    *pTCOUNT = (NumberOfCycles - 1)* ((VCO/VCO_CoreCLK)/TimerSlow);
    *pTSCALE = 0;
       
	*pTIMER6_WIDTH = SystemCLK / TimerSlow / 2;
	*pTIMER6_PERIOD = SystemCLK / TimerSlow;
	*pTIMER6_CONFIG = PERIOD_CNT | PWM_OUT;
	
	*pTIMER0_WIDTH = 2;
    *pTIMER0_CONFIG = PWM_OUT | TIN_SEL | CLK_SEL;
    
    *pTIMER_ENABLE = TIMEN6 | TIMEN0;
 	*pTCNTL |= TMREN;
 	*pTIMER_DISABLE = TIMDIS0;
}

void SetUpClockBFReceive()
{
    TimerInterruptCause = 2;
    ClockDone = 1;
	
    *pTIMER6_WIDTH = SystemCLK / TimerSlow / 2;
	*pTIMER6_PERIOD = SystemCLK / TimerSlow;
	*pTIMER6_CONFIG = PERIOD_CNT | PWM_OUT;
	
	*pTIMER_ENABLE = TIMEN6;
	ssync();
	*pTIMER_DISABLE = TIMDIS6;
	while(*pTIMER_STATUS & TRUN6);
	ClockDone = 0;
}

void SetUpClock(int NumberOfCycles)
{
    TimerInterruptCause = 3;
    ClockDone = 1;
    
    *pTIMER_DISABLE = TIMDIS7;
	while(*pTIMER_STATUS & TRUN7);
	
    *pTCNTL = TMPWR;
    *pTCOUNT = (NumberOfCycles-1)* ((VCO/VCO_CoreCLK)/TimerSlow);
    *pTSCALE = 0;
       
	*pTIMER6_WIDTH = SystemCLK / TimerSlow / 2;
	*pTIMER6_PERIOD = SystemCLK / TimerSlow;
	*pTIMER6_CONFIG = PERIOD_CNT | PWM_OUT;
	
	*pTIMER7_WIDTH = SystemCLK / TimerSlow / 2;
	*pTIMER7_PERIOD = SystemCLK / TimerSlow;
	*pTIMER7_CONFIG = PERIOD_CNT | PWM_OUT;

	*pTIMER_ENABLE = TIMEN6 | TIMEN7;
 	*pTCNTL |= TMREN;
}

void SetUpDMATransmit(int NumberOfWords)
{
    *pDMA0_START_ADDR = CommandBuffer[0];
    *pDMA0_X_COUNT = NumberOfWords;
    *pDMA0_X_MODIFY = 1;
    *pDMA0_CONFIG = FLOW_STOP | DMAEN | DI_EN; 
}

void SetUpDMADataTransmit(int NumberOfWords)
{
    *pDMA0_START_ADDR = TransmitData[0];
    *pDMA0_X_COUNT = NumberOfWords;
    *pDMA0_X_MODIFY = 1;
    *pDMA0_CONFIG = FLOW_STOP | DMAEN | DI_EN; 
}


void SetUpDMAReceive(int NumberOfWords)
{
    *pDMA0_START_ADDR = ResponseBuffer[0];
    *pDMA0_X_COUNT = NumberOfWords;
    *pDMA0_X_MODIFY = 1;
    *pDMA0_CONFIG = FLOW_STOP | WNR | DMAEN | DI_EN; 
}

void SetUpDMADataReceive(int NumberOfWords)
{
    *pDMA0_START_ADDR = ReceiveData[0];
    *pDMA0_X_COUNT = NumberOfWords;
    *pDMA0_X_MODIFY = 1;
    *pDMA0_CONFIG = FLOW_STOP | WNR | DMAEN | DI_EN; 
}

void SetUpAndEnablePPITransmit(int NumberOfWords)
{
    InterruptCause = 1;
    PPIDone = 1;
    *pPPI_DELAY = 0;
    *pPPI_COUNT = NumberOfWords;
    *pPPI_CONTROL = PORT_DIR | PORT_EN | POLS | XFR_TYPE;
}

void SetUpAndEnablePPIReceive()
{
    
    PPIDone = 1;
    *pPPI_CONTROL = XFR_TYPE | PORT_CFG | FLD_SEL | PORT_EN | POLS;
}

void SetUpCommandArray(int Command, int value)
{	
    int temp, temp1;
    char crc;
    int i,j,a,k,xor;
    char Data;
    
    for(i=0;i<6;i++)
	{
	   for(j=0;j<8;j++)
	   {
	       CommandBuffer[i][j] = 0;
	   }
	}
    // Filling up the command buffer
    CommandBuffer[0][0] = 0x40 | Command;
    CommandBuffer[1][0] = (value & 0xff000000) >> 24;
	CommandBuffer[2][0] = (value & 0xff0000) >> 16;
	CommandBuffer[3][0] = (value & 0xff00) >> 8;
	CommandBuffer[4][0] = value & 0xff;
	
	// CMD0 should be 0x40 00 00 00 00 95

	    crc=0; 
		for (a=0;a<5;a++) 
		{ 
			Data=CommandBuffer[a][0]; 
			for (k=0;k<8;k++) 
			{ 
				crc <<= 1; 

				if ((Data & 0x80)^(crc & 0x80)) 
				crc ^=0x09; 
				Data <<= 1; 
			} 
		} 
		crc=(crc<<1)|1; 
		CommandBuffer[5][0] = crc;
		
			
	for(i=0;i<6;i++)
	{
	   temp = CommandBuffer[i][0]; 
	   for(j=0;j<8;j++)
	   {
	       temp1 = (temp & (1<<j));
	       if(j<4)
	       {
	           CommandBuffer[i][7-j] = 0xF | (temp1 << (4-j));
	       }
	       if(j==4)
	       {
	           CommandBuffer[i][7-j] = 0xF | temp1; 
	       }
	       if(j>4)
	       {
	           CommandBuffer[i][7-j] = 0xF | (temp1 >> (j-4));
	       }
	   }
	}
}

void SetUpTransmitData()
{
	
	short crc, Data, temp;
	char DataIn;
	short a, b, c;
	int i, TempData;
	char j;
	
	TransmitData[0][0] = 0xFE;
        
    crc=0; 
    i=0;
	for (i=1;i<513;i++) 
	{
		for(j=0;j<8;j++)
		{
			DataIn = (TransmitData[i][0] & (1<<(7-j))) >> (7-j);
			a = (crc & 0x1)	^ DataIn;
			b = ((crc & 0x800)>>11) ^ a;
			c = ((crc & 0x10)>>4) ^ a;
			crc = crc >> 1;
			crc = crc & 0x7BF7;
			a = a << 15;
			b = b << 10;
			c = c << 3;
			crc = crc | (a & 0x8000);
			crc = crc | (b & 0x400);
			crc = crc | (c & 0x8);
		}
	}
	temp = 0;
	for(j=0;j<16;j++)
	{
		temp = temp | (((crc & (1<<j))>>j)<<(15-j));
	}
	
	TransmitData[513][0] = ((temp&0xFF00)>>8);
    TransmitData[514][0] = (temp&0xFF);
	
	for(i=0;i<515;i++)
	{
	   TempData = TransmitData[i][0]; 
	   for(j=0;j<8;j++)
	   {
	       TransmitData[i][j] = (0x1E | ((TempData & (1<<(7-j)))>>(7-j)));
	   }
	}
	TransmitData[515][0] = 0x1F;
}

void ExtractResponse()
{
    int i,j;
    for(i=0;i<17;i++)
    {
        FinalResponseBuffer[i] = 0;
    }
    for(i=0;i<17;i++)
    {
     	for(j=0;j<8;j++)
     	{
     	    ResponseBuffer[i][j] = ResponseBuffer[i][j] & (0x10);
     	    if(j<3)
     	    {
     	        ResponseBuffer[i][j] = ResponseBuffer[i][j] << (3-j);
     	    }
     	    if(j>3)
     	    {
     	        ResponseBuffer[i][j] = ResponseBuffer[i][j] >> (j-3);   	        
     	    }
     	    FinalResponseBuffer[i] |= ResponseBuffer[i][j];
     	}   
     }    
}

void ExtractData()
{
	int i,j;
    for(i=0;i<514;i++)
    {
        FinalReceiveData[i] = 0;
    }
    for(i=0;i<514;i++)
    {
     	for(j=0;j<8;j++)
     	{
     	    ReceiveData[i][j] = ReceiveData[i][j] & (0x1);
     	    ReceiveData[i][j] = ReceiveData[i][j] << (7-j);
     	    
     	    FinalReceiveData[i] |= ReceiveData[i][j];
     	}   
     }    
}

void ExtractRData()
{
	int i,j;
    for(i=0;i<514;i++)
    {
        FinalReceiveData[i] = 0;
    }
    for(i=0;i<514;i++)
    {
     	for(j=0;j<8;j++)
     	{
     	    ReceiveData[i][j] = ReceiveData[i][j] & (0x1);
     	    ReceiveData[i][j] = ReceiveData[i][j] << (7-j);
     	    
     	    FinalReceiveData[i] |= ReceiveData[i][j];
     	}   
     }    
     
     for(i=0;i<514;i++)
     {
     	FinalReceiveData[i] = FinalReceiveData[i+3];
     }
}

void SetupPFpin()
{
    *pPORTFIO_DIR = 0;
    *pPORTFIO_POLAR = PF9 | PF8;
    *pPORTFIO_EDGE = PF8 | PF9;
    *pPORTFIO_MASKA_SET = PF8;
    *pPORTFIO_MASKB_SET =  PF9;
    asm("nop;");asm("nop;");asm("nop;");asm("nop;");asm("nop;");asm("nop;");asm("nop;");
    asm("nop;");asm("nop;");asm("nop;");asm("nop;");asm("nop;");asm("nop;");asm("nop;");
    asm("nop;");asm("nop;");asm("nop;");asm("nop;");
    
}

void SetupPFInterrupt()
{
    *pILAT = EVT_IVG12;
    *pSIC_IMASK1 |= IRQ_PFA_PORTF; //Use this carefully!
    *pPORTFIO_INEN = PF8;
}

void SetupPFInterruptData()
{
    *pILAT = EVT_IVG13;
    *pSIC_IMASK1 |= IRQ_PFB_PORTF; //Use this carefully!
    *pPORTFIO_INEN = PF9;
}

void DisablePPITransmit()
{
   while(*pDMA0_IRQ_STATUS & DMA_RUN);
    *pPPI_CONTROL = 0;
    *pDMA0_CONFIG = 0x0;
}

void DisablePPIReceive()
{
    *pDMA0_CONFIG = 0x0;
    *pPPI_CONTROL = 0;
}

